#include "calculator.h"
#include <QPushButton>
#include <QLCDNumber>
#include <QGridLayout>
#include <QDebug>
#include <QObject>
#include <QRadioButton>
#include <QHBoxLayout>
#include <QtMath>


Calculator::Calculator(QWidget *parent) //constructor
    : QWidget(parent),
    f_lcdNum(0),
    f_currentNum(""),
    f_if_currNumChangable(true)
{
    setMinimumSize(500, 400);


    this->setWindowTitle("Calculator (standard)");

    MainLayout = new QGridLayout(this);
    f_BtnLayoutCasual = new QGridLayout(this);  //layout for casual calculator buttons
    f_BtnLayoutSciExtr = new QGridLayout(this); //layout for scientific calculator buttons
    f_BtnLayoutSciExtrWidget = new QWidget(this);   //widget for sci buttons
    f_BtnLayoutCasualExtrWidget = new QWidget(this);    //widget for casual buttons
    f_HBoxBtnLayout = new QHBoxLayout(this);    //box layout for all buttons

    //QLCDNumber initialization
    f_lcdNum = new QLCDNumber(12, this);
    f_lcdNum->setSegmentStyle(QLCDNumber::Flat);
    f_lcdNum->setMinimumSize(370,60);
    f_lcdNum->setDigitCount(20);

    //QRadioButtons initialization
    f_radioBtnCasual = new QRadioButton("Standard", this);
    f_radioBtnCasual->toggle();
    f_radioBtnSci = new QRadioButton("Scientific", this);

    connect(f_radioBtnCasual, SIGNAL(toggled(bool)), this, SLOT(slotHideSci(bool)));
    connect(f_radioBtnSci, SIGNAL(toggled(bool)), this, SLOT(slotShowSci(bool)));

    //Button layouts initialization
    QString ButtonsTxtCasual[4][5] = {
        {"7","8","9","/","sqrt"},
        {"4","5","6","*","rev"},
        {"1","2","3","-","sqr"},
        {".","0","=","+","-x"}
    };

    QString ButtonsTxtSciExtr[4][4] = {
        {"sinh", "sin", "exp", "ln"},
        {"cosh", "cos", "n!", "log"},
        {"tanh", "tan", "cubic sqr", "ysqr"},
        {"pi", "e", "cubic root", "ysqrt"}
    };



    for (int row = 0; row < 4; row++)
    {
        for (int col = 0; col < 5; col++)
        {
            f_BtnLayoutCasual->addWidget(createButton(ButtonsTxtCasual[row][col], 80, 60), row+2, col);
        }
    }



    for (int row = 0; row < 4; row++)
    {
        for (int col = 0; col < 4; col++)
        {
            f_BtnLayoutSciExtr->addWidget(createButton(ButtonsTxtSciExtr[row][col], 80, 60), row+2, col);
        }
    }

    f_BtnLayoutSciExtrWidget->setLayout(f_BtnLayoutSciExtr);
    f_BtnLayoutCasualExtrWidget->setLayout(f_BtnLayoutCasual);


    //Button box initialization

    f_HBoxBtnLayout->addWidget(f_BtnLayoutSciExtrWidget);
    f_HBoxBtnLayout->addWidget(f_BtnLayoutCasualExtrWidget);


    //Main layout initialization
    MainLayout->addWidget(f_lcdNum, 0, 0, 1, 8, Qt::AlignmentFlag::AlignHCenter);
    MainLayout->addWidget(createButton("CE", 80, 60), 0, 8, Qt::AlignmentFlag::AlignCenter);

    MainLayout->addWidget(f_radioBtnCasual, 1, 0, 1, 2, Qt::AlignmentFlag::AlignCenter);
    MainLayout->addWidget(f_radioBtnSci, 1, 2, 1, 2, Qt::AlignmentFlag::AlignCenter);

    MainLayout->addLayout(f_HBoxBtnLayout, 2, 0, 4, 9, Qt::AlignmentFlag::AlignCenter);


    f_BtnLayoutSciExtrWidget->hide();

    this->setLayout(MainLayout);
    resize (500, 400);
}


//Button creation method
QPushButton* Calculator::createButton(const QString &p_str, size_t p_horz, size_t p_vert)
{
    QPushButton* new_PBtn = new QPushButton(p_str);
    new_PBtn->setMinimumSize(p_horz, p_vert);
    connect(new_PBtn, SIGNAL(clicked()), this, SLOT(slotButtonClicked()));
    return new_PBtn;
}

Calculator::~Calculator()   //destructor
{

}

void Calculator::calculate()
{
    if (f_operStack.size() < 2)
        return;

    if (f_operStack.size() == 2)
    {
        QString operation = f_operStack.pop();
        double operand = f_operStack.pop().toDouble();
        qDebug() << operand << operation << " 0 if binary";

        if (operation == "rev")
        {
            if (operand == 0)
            {
                f_lcdNum->display("ERR");
                qDebug() << "ERR";
            }
            else
                f_lcdNum->display(1 / operand);
            qDebug() << 1/operand;
        }

        if (operation == "sqr")
        {f_lcdNum->display(operand * operand);
        qDebug() << operand * operand;}

        if (operation == "sqrt")
        {f_lcdNum->display(sqrt(operand));
        qDebug() << sqrt(operand);}

        if(operation == "-x")
        {f_lcdNum->display(-operand);
        qDebug() << -operand;}

        if(operation == "sin")
        {f_lcdNum->display(qSin(operand));
        qDebug() << qSin(operand);}

        if(operation == "sinh")
        {f_lcdNum->display(qAsin(operand));
        qDebug() << qAsin(operand);}

        if(operation == "cos")
        {f_lcdNum->display(qCos(operand));
        qDebug() << qCos(operand);}

        if(operation == "cosh")
        {f_lcdNum->display(qAcos(operand));
        qDebug() << qAcos(operand);}

        if(operation == "tan")
        {f_lcdNum->display(qTan(operand));
        qDebug() << qTan(operand);}

        if(operation == "tanh")
        {f_lcdNum->display(qAtan(operand));
        qDebug() << qAtan(operand);}

        if(operation == "ln")
        {f_lcdNum->display(qLn(operand));
        qDebug() << qLn(operand);}

        if(operation == "exp")
        {f_lcdNum->display(qExp(operand));
        qDebug() << qExp(operand);}

        if(operation == "cubic root")
        {   double tt = 1/3.0;
            f_lcdNum->display(qPow(operand, tt));
        qDebug() << qPow(operand, tt);}

        if(operation == "cubic sqr")
        {f_lcdNum->display(operand * operand * operand);
        qDebug() << operand * operand * operand;}

        if(operation == "n!")
        {
            int res = 1;
            int oper = static_cast<int>(operand);
            if (oper < 1)
            {
                f_lcdNum->display("Err");
                return;
            }
            else
            {
                while (oper > 1)
                {
                    res *= oper;
                    oper--;
                }
            }
            f_lcdNum->display(res);
        qDebug() << res;}

        if (operation == "+" || operation == "-")
        {f_lcdNum->display(operand);
        qDebug() << operand;}

        if (operation == "*")
        {f_lcdNum->display(0);
        qDebug() << 0;}

        if (operation == "/")
        {f_lcdNum->display("Err");
        qDebug() << "Err";}

        if (operation == "log")
        {f_lcdNum->display("Err");
        qDebug() << "Err";}

        if (operation == "ysqr")
        {f_lcdNum->display(1);
        qDebug() << 1;}

        if (operation == "ysqrt")
        {f_lcdNum->display("Err");
        qDebug() << "Err";}


    }
    else if(f_operStack.size() == 3)
    {

        double operand2 = f_operStack.pop().toDouble();
        QString operation = f_operStack.pop();
        double operand1 = f_operStack.pop().toDouble();

        qDebug() << operand1 << operation << operand2 << ":";

        if (operation == "+")
        {f_lcdNum->display(operand1 + operand2);
        qDebug() << operand1 + operand2;}

        if (operation == "-")
        {f_lcdNum->display(operand1 - operand2);
        qDebug() << operand1 - operand2;}

        if (operation == "*")
        {f_lcdNum->display(operand1 * operand2);
        qDebug() << operand1 * operand2;}

        if (operation == "/")
        {f_lcdNum->display(operand1 / operand2);
        qDebug() << operand1 / operand2;}

        if (operation == "log")
        {f_lcdNum->display(qLn(operand1) / qLn(operand2));
        qDebug() << qLn(operand1) / qLn(operand2);}

        if (operation == "ysqr")
        {f_lcdNum->display(qPow(operand1, operand2));
        qDebug() << qPow(operand1, operand2);}

        if (operation == "ysqrt")
        {f_lcdNum->display(qPow(operand1, 1.0 / operand2));
        qDebug() << qPow(operand1, 1.0 / operand2);}
    }
    else if (f_operStack.size() == 5)
    {   double operand3 = f_operStack.pop().toDouble();
        QString operation2 = f_operStack.pop();
        double operand2 = f_operStack.pop().toDouble();
        QString operation1 = f_operStack.pop();
        double operand1 = f_operStack.pop().toDouble();
        double mid_res;

        qDebug() << operand1 << operation1 << operand2 << operation2 << operand3 << ":";



        if (operation2 == "*")
        {mid_res = operand2 * operand3;
        qDebug() << "mid_res = " << operand2 * operand3;}

        if (operation2 == "/")
        {mid_res = operand2 / operand3;
        qDebug() << "mid_res = " << operand2 / operand3;}

        if (operation2 == "log")
        {mid_res = qLn(operand2) / qLn(operand3);
        qDebug() << "mid_res = " << qLn(operand2) / qLn(operand3);}

        if (operation2 == "ysqr")
        {mid_res = qPow(operand2, operand3);
        qDebug() << "mid_res = " << qPow(operand2, operand3);}

        if (operation2 == "ysqrt")
        {mid_res = qPow(operand2, 1.0 / operand3);
        qDebug() << "mid_res = " << qPow(operand2, 1.0 / operand3);}



        if (operation1 == "+")
        {f_lcdNum->display(operand1 + mid_res);
        qDebug() << operand1 + mid_res;}

        if (operation1 == "-")
        {f_lcdNum->display(operand1 - mid_res);
        qDebug() << operand1 - mid_res;}

        if (operation1 == "*")
        {f_lcdNum->display(operand1 * mid_res);
        qDebug() << operand1 * mid_res;}

        if (operation1 == "/")
        {f_lcdNum->display(operand1 / mid_res);
        qDebug() << operand1 / mid_res;}

        if (operation1 == "log")
        {f_lcdNum->display(qLn(operand1) / qLn(mid_res));
        qDebug() << qLn(operand1) / qLn(mid_res);}

        if (operation1 == "ysqr")
        {f_lcdNum->display(qPow(operand1, mid_res));
        qDebug() << qPow(operand1, mid_res);}

        if (operation1 == "ysqrt")
        {f_lcdNum->display(qPow(operand1, 1.0 / mid_res));
        qDebug() << qPow(operand1, 1.0 / mid_res);}


    }
}

//slots: public

void Calculator::slotHideSci(bool p_is)
{
    if (p_is)
    {
        this->setWindowTitle("Calculator (standard)");
        f_BtnLayoutSciExtrWidget->hide();

        setMinimumSize(500, 400);
        resize(500, 400);

        qDebug() << "Resize (500, 400)";
        f_lcdNum->setMinimumSize(370,60);
        f_lcdNum->setDigitCount(20);
    }
}

void Calculator::slotShowSci(bool p_is)
{
    if (p_is)
    {
        this->setWindowTitle("Calculator (scientific)");
        f_BtnLayoutSciExtrWidget->show();

        setMinimumSize(850, 400);
        resize(850, 400);

        qDebug() << "Resize (850, 400)";
        f_lcdNum->setMinimumSize(650,60);
        f_lcdNum->setDigitCount(35);
    }
}

void Calculator::slotButtonClicked()
{
    QString input = dynamic_cast<QPushButton*>(sender())->text();
    //!!! Где проверка прапвильности приведения? sender не обязательно QPushButton
    
    
    qDebug() << "Just pressed: " << input;

    if (input == "CE")  //if CE: clear stack, display 0
    {
        f_operStack.clear();
        f_currentNum = "";
        f_lcdNum->display(0);
        return;
    }

    if (input == "C")   //no button on GUI, however realization is like this:
    {
        if (f_currentNum != "") //if there is currently inputed number, delete it, save stack
        {
            f_currentNum = "";
            f_lcdNum->display(0);
            return;
        }
        else    //if not, clear stack
        {
            f_operStack.clear();
            f_lcdNum->display(0);
            return;
        }
    }

    if (input.contains(QRegExp("[0-9]")))    //if number:
    {
        if (f_if_currNumChangable)  //if current number is not a numerical constant
        {
            if (input == "0" && f_currentNum == "") //dispay 0, leave currentNum empty if inputed inintial 0
            {
                f_lcdNum->display(0);
                return;
            }
            f_currentNum += input;  //else add and display
            f_lcdNum->display(f_currentNum);
            return;
        }
        else    //if current number is a numerical constant
        {
            f_if_currNumChangable = true;   //make sure it will be changable after clearing
            if (input == "0")   //if 0
            {
                f_currentNum = "";  //make current num resemble 0
                f_lcdNum->display(0);   //display 0
                return;
            }
            f_currentNum = input;   //else displayinputed number
            f_lcdNum->display(f_currentNum);
            return;
        }
    }
    else if (input == ".")           //if dot
    {
        if (f_currentNum == "")             //check if current number is empty
        {
            f_currentNum += "0.";   //if is, add initial 0 with dot and display
            f_lcdNum->display(f_currentNum);
            return;
        }
        else    //if not empty
        {
            if (f_if_currNumChangable)  //if is changable
            {
                if (f_currentNum.contains('.')) //if there is a dot already return, or start a new number over
                {
                    /*f_currentNum = "0.";
                    f_lcdNum->display(f_currentNum);
                    return;*/
                    return;
                }
                f_currentNum += input;          //if not: add, display
                f_lcdNum->display(f_currentNum);
                return;
            }
            else    //if not changable
            {
                f_if_currNumChangable = true;   //make sure it will be changable after clearing
                f_currentNum = "0.";    //add initial 0 with dot and display
                f_lcdNum->display(f_currentNum);
                return;
            }
         }
    }
    else if (NumND.contains(input)) //if numerical constant
    {
        if (input == "e")
        {
            f_currentNum = QString::number(qExp(1));
            f_if_currNumChangable = false;      //make sure to make it not changable not to add new numbers to numerical constant
            f_lcdNum->display(f_currentNum);
            return;
        }
        else
        {
            f_currentNum = QString::number(PI);
            f_if_currNumChangable = false;      //make sure to make it not changable not to add new numbers to numerical constant
            f_lcdNum->display(f_currentNum);
            return;
        }
    }
    else        //if input is an operation
    {
        if (input == "=")   //if =
        {
            if (f_operStack.size() == 0)    //if nothing to calculate, currently inputed number is already displayed
            {
                return;
            }
            else if (f_operStack.size() == 2)   //if 1 operand and binary operation already inputed
            {
                if (f_currentNum == "") //if no second operand
                {
                    calculate();    //calculate with second operand as 0
                    return;
                }
                else    //if there is a second operand
                {
                    f_operStack.push_back(f_currentNum);    //add it to stack and calculate, clear current number
                    calculate();
                    f_operStack.clear();
                    f_currentNum = "";
                    return;
                }
            }
            else    //if 2 operands and 2 operations already inputed
            {
                if (f_currentNum == "") //if last (3-rd) operand is empty
                {
                    f_operStack.pop_back(); //drop second operator
                    calculate();    //calculate single operation
                    f_operStack.clear();
                    return;
                }
                else    //if last (3-rd) operand is not empty
                {
                    f_operStack.push_back(f_currentNum);   //add it to stack, calculate
                    f_currentNum = "";
                    calculate();
                    f_operStack.clear();
                    return;
                }
            }
        }
        else    //if operation
        {
            if (f_operStack.size() == 0)    //if number 1 is being inputed at the time
            {
                if (f_currentNum == "") //if none already inputed
                {
                    f_operStack.push_back(QString::number(f_lcdNum->value()));  //get result of last operation or initial 0 from lcd number
                    f_operStack.push_back(input);   //push back inputed operation
                    if (UnOp.contains(input))   //if unary operation just inputed -- calculate immediately
                    {
                        calculate();
                        f_operStack.clear();
                        return;
                    }
                    return;
                }
                else    //if number already inputed
                {
                    f_operStack.push_back(f_currentNum);
                    f_currentNum = "";      //push back inputed number and calculate immediately if operation just inputed is unary
                    f_operStack.push_back(input);
                    if (UnOp.contains(input))
                    {
                        calculate();
                        f_operStack.clear();
                        return;
                    }
                    return;
                }
            }
            else if (f_operStack.size() == 2) //if number 2 is being inputed at the time, operation stack contains number 1 and binary operation
            {
                if (f_currentNum == "") //if number 2 is empty
                {
                    f_operStack.pop_back(); //change binary operation in stack to inputed, calculate if it is unary
                    f_operStack.push_back(input);
                    if (UnOp.contains(input))
                    {
                        calculate();
                        f_operStack.clear();
                        return;
                    }
                    return;
                }
                else    //if number 2 is not empty
                {

                    QString tech_oper;      //get previous operation
                    tech_oper = f_operStack.pop();

                    if (OpToCountimmediately.contains(tech_oper) || SumOp.contains(input))  //if inputed +, - or the first operator is a binary operator wich is supposed to counted immediately
                    {
                        f_operStack.push_back(tech_oper);
                        f_operStack.push_back(f_currentNum);    //add currenty inputed number
                        f_currentNum = "";
                        calculate();    //calculate with previously inputed operation
                        f_operStack.clear();
                        f_operStack.push_back(QString::number(f_lcdNum->value()));  //get result, add it to operation stack
                        f_operStack.push_back(input);   //add + or - to operation stack
                        return;
                    }
                    else
                    {
                        if (UnOp.contains(input))   //if inputed unary operation    (*)
                        {
                            QString tech_f_operand; //get first number
                            tech_f_operand = f_operStack.pop();

                            f_operStack.push_back(f_currentNum);    //calculate second number with inputed unary operation
                            f_operStack.push_back(input);
                            calculate();
                            f_operStack.clear();
                            f_operStack.push_back(tech_f_operand);  //push first operand and operation back
                            f_operStack.push_back(tech_oper);
                            f_currentNum = QString::number(f_lcdNum->value());  //make result a currently inputed number
                            f_if_currNumChangable = false; //make it unchangable as it is a result of unary operation
                            return;
                        }
                        else    //else add second number and second operation to stack
                        {
                            f_operStack.push_back(f_currentNum);
                            f_currentNum = "";
                            f_operStack.push_back(input);
                            return;
                        }
                    }
                }
            }
            else    //if number 3 is being inputed at the time and operation stack contains number 1, binary operation 1, number 2, binary operation 2
            {
                if (f_currentNum == "") //if number 3 is empty
                {
                    f_operStack.pop_back(); //drop operation 2
                    if (UnOp.contains(input))   //if newly inputed operation 2 is unary, do the same as in (*)
                    {
                        QString tech_f_operand;
                        QString tech_s_operand;
                        QString tech_oper;
                        tech_s_operand = f_operStack.pop();
                        tech_oper = f_operStack.pop();
                        tech_f_operand = f_operStack.pop();
                        f_operStack.push_back(tech_s_operand);
                        f_operStack.push_back(input);
                        calculate();
                        f_operStack.clear();
                        f_operStack.push_back(tech_f_operand);
                        f_operStack.push_back(tech_oper);
                        f_currentNum = QString::number(f_lcdNum->value());
                        return;
                    }
                    else    //if newly inputed operation 2 is binary, add it to stack in place of old binary operation
                    {
                        f_operStack.push_back(input);
                        return;
                    }
                }
                else    //if number 3 is not empty
                {
                    if (UnOp.contains(input))   //if inputed operation 3 is unary
                    {
                        QString tech_f_operand;
                        QString tech_s_operand;
                        QString tech_oper1;
                        QString tech_oper2;

                        tech_oper2 = f_operStack.pop();
                        tech_s_operand = f_operStack.pop();     //get previously inputed numbers and operations out of stack
                        tech_oper1 = f_operStack.pop();
                        tech_f_operand = f_operStack.pop();

                        f_operStack.push_back(f_currentNum);    //calculate number 3 and inputed unaru operation
                        f_operStack.push_back(input);
                        calculate();
                        f_operStack.clear();
                        f_currentNum = QString::number(f_lcdNum->value());  //make current number a result of calculation
                        f_if_currNumChangable = false;  //make it unchangable as it is a result of unary operation

                        f_operStack.push_back(tech_f_operand);
                        f_operStack.push_back(tech_oper1);  //push previously inputed numbers and operations to stack
                        f_operStack.push_back(tech_s_operand);
                        f_operStack.push_back(tech_oper2);
                        return;
                    }
                    else    //if inputed operation 3 is binary
                    {
                        if (SumOp.contains(input))  //if it is + or -
                        {
                            f_operStack.push_back(f_currentNum);    //calculate previous complicated stack
                            f_currentNum = "";
                            calculate();
                            f_operStack.clear();
                            f_operStack.push_back(QString::number(f_lcdNum->value()));  //push the result to stack
                            f_operStack.push_back(input);   //add + or - to stack as operation 1
                            return;
                        }
                        QString tech_f_operand; //else get previously inputed numbers and operations out of stack
                        QString tech_s_operand;
                        QString tech_oper1;
                        QString tech_oper2;

                        tech_oper2 = f_operStack.pop();
                        tech_s_operand = f_operStack.pop();
                        tech_oper1 = f_operStack.pop();
                        tech_f_operand = f_operStack.pop();

                        f_operStack.push_back(tech_s_operand);  //calculate [number 2 <operation 2> number 3]
                        f_operStack.push_back(tech_oper2);
                        f_operStack.push_back(f_currentNum);
                        calculate();
                        f_operStack.clear();
                        f_operStack.push_back(tech_f_operand);  //push number 1, operation 1, result and newly inputed operation back to stack
                        f_operStack.push_back(tech_oper1);
                        f_operStack.push_back(QString::number(f_lcdNum->value()));
                        f_operStack.push_back(input);
                        f_currentNum = "";
                        return;
                    }
                }
            }
        }
    }
}
